/*
 * License for Tiled Display modifications
 *
 * Copyright (c) 2005 Mohamed Ibrahim and Stephanie Smullen.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include an acknowlegement that the software contains
 *    code based on contributions made by Mohamed Ibrahim and Stephanie Smullen and
 *    include a link to http://zog.utc.edu/~vislab/.
 *
 * 4.  The Original code contains third party software libraries copyrighted 
 *	under their own licenses. Redistribution of Libraries in source code, 
 *	with or without modification, or redistribution in binary form falls
 *	under the terms of the licenses for the corresponding Libraries. 
 *
 * This software is provided "as is" and any expressed or implied
 * warranties, including, but not limited to, the implied warranties
 * of merchantability and fitness for a particular purpose are
 * disclaimed.  In no event shall the authors or their
 * employers be liable for any direct, indirect, incidental,
 * special, exemplary, or consequential damages (including, but not
 * limited to, procurement of substitute goods or services; loss of
 * use, data, or profits; or business interruption) however caused and
 * on any theory of liability, whether in contract, strict liability,
 * or tort (including negligence or otherwise) arising in any way out
 * of the use of this software, even if advised of the possibility of
 * such damage.
 *

 *
 */
/**
 * Original code: Brian Dudley [Brian-Dudley at utc.edu],  June 2004
 * Modifications: Mohamed Ibrahim [Mohamed-Ibrahim at utc.edu], September 2004
 */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
/* Makes pngtest verbose so we can find problems (needs to be before png.h) */
#ifndef PNG_DEBUG
#  define PNG_DEBUG 0
#endif
#include "png.h"		// for PNG support
extern "C"
{
#include "jpeglib.h"	// for JPEG support
}
#include "jerror.h"
#include "jconfig.h"	
#include "jmorecfg.h"
#include <setjmp.h>
extern "C"
{
#include "gif_lib.h"	// for GIF support
#include "getarg.h"
}
#include "mpi_opengl.h"
#include "file_handler.h"
#include "network_comm.h"

#include <fcntl.h>		// GIF file control options
#include <ctype.h>


TEXTURE *read_file (char *infilename) {
    strlwr(infilename);
    if (strstr(infilename, ".ppm") != 0)
        return read_ppm(infilename);
    else if (strstr(infilename, ".bmp") != 0)
        return read_bmp(infilename); 
    else if (strstr(infilename, ".png") != 0)
        return read_png(infilename);	 
    else if (strstr(infilename, ".jpg") != 0)
        return read_jpg(infilename);
    else if (strstr(infilename, ".gif") != 0)
        return read_gif(infilename);
    else {
        debug("Only PPM, BMP, PNG, JPG, and GIF file formats are currently supported.\n");
        return NULL;
    }
}

// Thanks to Russ Burdick for die,debug, the TEXTURE structure and 
// the basic read_texture file (renamed read_ppm here for clarity)
// <wburdick@cs.umn.edu>
// Modification to read_ppm made by Clint Smullen,IV, Clemson University
// are indicated by //******
// they allow more than 255 image max and use rb rather than r mode


//**************************************read_ppm*******************************
TEXTURE* read_ppm(char *infilename) {
    
    FILE *infile;
    char string[STRING_SIZE];
    int i, j, k, l;
    int r, g, b;
    int img_max, img_width, img_height;
    TEXTURE *tex;
    unsigned char *ibuffer;
    int ascii = 0;
    
    /* open file containing texture */
    if ((infile = fopen(infilename, "rb")) == NULL) {
        die("read_ppm: error opening texture file %s\n", infilename);
    }
    
    /* read and discard first two lines */
    fgets(string, STRING_SIZE, infile);
    if (strncmp(string, "P3", 2) == 0) { /* ascii ppm */
        debug("read_ppm: found ascii ppm \"%s\"\n", infilename);
        ascii = 1;
    } else if (strncmp(string, "P6", 2) == 0) { /* raw ppm */
        debug("read_ppm: found raw ppm \"%s\"\n", infilename);
        ascii = 0;
    } else {
        die("read_texture: image file \"%s\" doesnt look like a ppm\n", infilename);
    }
    
    fgets(string, STRING_SIZE, infile);
    while (!feof(infile) && (string[0] == '#')) {
        fgets(string, STRING_SIZE, infile);
    }
    
    /* read image size and (?)max component value */
    sscanf(string, "%d %d", &img_width, &img_height);
    debug("read_ppm: ppm \"%s\" is %dx%d\n", infilename, img_width,
        img_height);
    
    fgets(string, STRING_SIZE, infile);
    sscanf(string, "%d", &img_max);
    debug("read_ppm: ppm \"%s\" has img_max = %d\n", infilename,
        img_max);
    
    /* allocate texture array */
    if ((tex = (TEXTURE *)malloc(sizeof(TEXTURE))) == NULL) {
        die("read_ppm: error allocating texture image");
    }
    
    ibuffer = (unsigned char *) malloc(img_height*img_width*(RGBA-1));
    if (ibuffer == NULL) die("read_ppm: error mallocing buffer\n");
    
    tex->width = img_width;
    tex->height = img_height;
    tex->thumb = NULL;
    tex->tex = (GLubyte *)malloc(img_height*img_width*RGBA);
    //
    if (ascii) { /* ascii ppm */
        
        /* read image data */
        for (i=0; i<img_height; i++) {
            for (j=0; j<img_width; j++) {
                fscanf(infile, "%d %d %d", &r, &g, &b);
                *(tex->tex + (RGBA*((img_height-1-i)*img_width+j)+0)) =
                    (GLubyte) (((int)r*255)/img_max);  // allow non-255 image max by scaling
                *(tex->tex + (RGBA*((img_height-1-i)*img_width+j)+1)) =
                    (GLubyte) (((int)g*255)/img_max);  // allow non-255 image max
                *(tex->tex + (RGBA*((img_height-1-i)*img_width+j)+2)) =
                    (GLubyte) (((int)b*255)/img_max);  // allow non-255 image max
                *(tex->tex + (RGBA*((img_height-1-i)*img_width+j)+3)) =
                    (GLubyte) 255;						// set alpha bit
            }
        }
        
    } else { /* raw ppm */
        
        fread(ibuffer, sizeof(unsigned char), img_height*img_width*(RGBA-1),infile);
        fclose(infile);
        
        l = 0;
        /* read image data */
        for (i=0; i<img_height; i++) {
            for (j=0; j<img_width; j++) {
                for (k = 0; k < RGBA-1; k++) 
                    *(tex->tex + (RGBA*((img_height-1-i)*img_width+j)+k)) =
                    (GLubyte) (((int)ibuffer[l++]*255)/img_max);  // allow non-255 image max
                *(tex->tex + (RGBA*((tex->height-1-i)*tex->width+j)+3)) = 255;
            }
        }
    }
    free(ibuffer);
    
    return tex;
    
}

//*********************************read_png*******************************************
// Thanks to Clint Smullen, IV -- Clemson University for his many helpful corrections
// and suggestion with the png files
TEXTURE* read_png(char *infilename) {
    FILE *infile;
    TEXTURE *tex;
    int number = 8, y;
    unsigned char header[8];
    
    png_structp png_ptr;
    png_infop info_ptr,end_info;
    
    // open file
    if(!(infile = fopen(infilename, "rb")))
        return 0;
    
    fread(header, 1, number, infile);
    bool is_png = !png_sig_cmp(header, 0, number);
    if (!is_png)
    {
        debug ("read_png: Error: File is not png type.");
    }
    debug("read_png: Opened png file OK.\n");
    
    // allocate and initialize png_struct and png_info
    if(!(png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, infilename, NULL, NULL))) {
        fclose(infile);
        return 0;
    }
    
    if(!(info_ptr = png_create_info_struct(png_ptr))) {
        png_destroy_read_struct(&png_ptr, NULL, NULL);
        fclose(infile);
        return 0;
    }
    
    if(!(end_info = png_create_info_struct(png_ptr))) {
        png_destroy_read_struct(&png_ptr,&info_ptr,NULL);
        fclose(infile);
        return 0;
    }
    
    if(setjmp(png_jmpbuf(png_ptr))) {
        png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
        fclose(infile);
        return 0;
    }
    debug("read_png: Allocated structs OK.\n");
    
    // setup the input code
    png_init_io(png_ptr, infile);
    png_set_sig_bytes(png_ptr, number); // some bytes missing
    
    png_read_info(png_ptr, info_ptr);  // read file info up to image data
    debug("read_png: Read file up to image data OK.\n");
    
    png_uint_32 width, height;			// define query variables
    int bit_depth, color_type, interlace_type;
    int compression_type, filter_method;
    // query info structure
    png_get_IHDR(png_ptr, info_ptr, 
        &width, &height,
        &bit_depth, &color_type, &interlace_type,
        &compression_type, &filter_method);
    debug("read_png: Queried info structure OK.\n");
    debug("read_png: png \"%s\" is %dx%d with bit depth= %d and color type= %d\n", infilename, width,height,bit_depth, color_type);
    
    // Set to a single byte, multibyte storage order	
    // extract multiple pixels from single byte to separate bytes
    if (bit_depth < 8)
        png_set_packing(png_ptr);
    
    // change storage to little endian
    if (bit_depth > 8)
    {
        #ifndef ITK_WORDS_BIGENDIAN
        png_set_swap(png_ptr);
        #endif
    }
    
    // Convert each color type to RGBA
    // minimum of a byte per pixel, then convert 1 byte to RGBA 
    if (color_type == PNG_COLOR_TYPE_GRAY) 
    {
        debug("read_png: Color type GRAY converted to 8 bit, rgb, alpha\n");
        //Up to 8 bits if 1, 2 or 4
        if (bit_depth <8)
            png_set_gray_1_2_4_to_8(png_ptr);
        //Downscale the 16-bit input channel to 8-bit
        if (bit_depth == 16)
            png_set_strip_16(png_ptr);
        //Change gray to rgb
        png_set_gray_to_rgb(png_ptr);
        // pad with alpha = 255
        png_set_filler (png_ptr, 0xff, PNG_FILLER_AFTER);
    }
    
    // minimum of a byte per pixel, then convert 1 byte to RGB and add alpha 
    if (color_type == PNG_COLOR_TYPE_GRAY_ALPHA) 
    {
        debug("read_png: Color type GRAY ALPHA converted to 8 bit, rgb, alpha\n");
        //Up to 8 bits if 1, 2 or 4
        if (bit_depth <8)
            png_set_gray_1_2_4_to_8(png_ptr);
        //Downscale the 16-bit input channel to 8-bit
        if (bit_depth == 16)
            png_set_strip_16(png_ptr);
        //Change gray to rgb
        png_set_gray_to_rgb(png_ptr);
        // add alpha if any alpha found
        if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS)) 
        {
            png_set_tRNS_to_alpha(png_ptr);
        }
        //or pad with alpha = 255
        else
            png_set_filler (png_ptr, 0xff, PNG_FILLER_AFTER);
    }
    
    // minimum of a byte per pixel, then convert palettes to RGBA
    if (color_type == PNG_COLOR_TYPE_PALETTE)
    {	debug("read_png: Color type PALETTE converted to 8 bit, rgb, alpha\n");
    //Up to 8 bits if 1, 2 or 4
    if (bit_depth <8)
        png_set_gray_1_2_4_to_8(png_ptr);
    //Change palette to rgb
    png_set_palette_to_rgb(png_ptr);
    // pad with alpha = 255
    png_set_filler (png_ptr, 0xff, PNG_FILLER_AFTER);
    }
    
    // Convert 3 byte RGB to 4 byte color with 4th byte = 255
    if (color_type = PNG_COLOR_TYPE_RGB)
    {	debug("read_png: Color type RGB converted to 8 bit, rbg, alpha\n");
    //Downscale the 16-bit input channel to 8-bit
    if (bit_depth == 16)
        png_set_strip_16(png_ptr);
    png_set_filler (png_ptr, 0xff, PNG_FILLER_AFTER);
    }
    
    // Convert 16 bit RGBA to 8 bit RGBA
    if (color_type = PNG_COLOR_TYPE_RGB_ALPHA)
    {	debug("read_png: Color type RGBA converted to 8 bit, rgb, alpha\n");
    //Downscale the 16-bit input channel to 8-bit
    if (bit_depth == 16)
        png_set_strip_16(png_ptr);
    // add alpha if any alpha found
    if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS)) 
    {
        png_set_tRNS_to_alpha(png_ptr);
    }
    //or pad with alpha = 255
    else
        png_set_filler (png_ptr, 0xff, PNG_FILLER_AFTER);
    }
    
    
    // update the info now that we have defined the filters
    png_read_update_info(png_ptr, info_ptr);
    
    debug("read_png: Input transformations OK.\n");
    
    // allocate texture array 
    if ((tex = (TEXTURE *)malloc(sizeof(TEXTURE))) == NULL) {
        die("read_png: error allocating texture image");
    }
    
    tex->width = width;
    tex->height = height;
    tex->thumb = NULL;
    tex->tex=(GLubyte *)malloc(height*width*RGBA);
    debug("read_png: Allocate texture array OK.\n");
    
    png_bytep *row_pointers = (png_bytep*) malloc(sizeof(png_bytep) * height);
    
    //-Clint Smullen, IV Clemson University - Points into the actual image data,
    // rather than allocating new buffers.
    for(y=0; y<(int)height; y++)
        row_pointers[y] = tex->tex+(height-y-1)*tex->width*RGBA;
    png_read_image(png_ptr, row_pointers);
    debug("read_png: Read image OK.\n");
    
    free(row_pointers);
    
    // close the file
    png_read_end(png_ptr, NULL);
    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
    
    debug("read_png: Close file OK.\n");
    
    return tex;
}

//**********************************read_jpg********************************************
struct my_error_mgr {
    struct jpeg_error_mgr pub;	// "public" fields 
    jmp_buf setjmp_buffer;		// for return to caller 
};

typedef struct my_error_mgr * my_error_ptr;

TEXTURE* read_jpg(char *infilename) {
    
    TEXTURE *tex;
    int number = 10;
    
    /* This struct contains the JPEG decompression parameters and pointers to
    * working space (which is allocated as needed by the JPEG library).
    */
    struct jpeg_decompress_struct cinfo;
    /* We use our private extension JPEG error handler.
    * Note that this struct must live as long as the main JPEG parameter
    * struct, to avoid dangling-pointer problems.
    */
    struct my_error_mgr jerr;
    /* More stuff */
    FILE * infile;		/* source file */
    JSAMPARRAY buffer;		/* Output row buffer */
    int row_stride;		/* physical row width in output buffer */
    
    
                        /* In this example we want to open the input file before doing anything else,
                        * so that the setjmp() error recovery below can assume the file is open.
                        * VERY IMPORTANT: use "b" option to fopen() if you are on a machine that
                        * requires it in order to read binary files.
    */
    
    if ((infile = fopen(infilename, "rb")) == NULL) {
        debug("read_jpg: can't open %s\n", infilename);
        return 0;
    }
    
    /* Step 1: allocate and initialize JPEG decompression object */
    
    /* We set up the normal JPEG error routines, then override error_exit. */
    cinfo.err = jpeg_std_error(&jerr.pub);
    // jerr.pub.error_exit = my_error_exit;
    /* Establish the setjmp return context for my_error_exit to use. */
    if (setjmp(jerr.setjmp_buffer)) {
    /* If we get here, the JPEG code has signaled an error.
    * We need to clean up the JPEG object, close the input file, and return.
        */
        jpeg_destroy_decompress(&cinfo);
        fclose(infile);
        return 0;
    }
    /* Now we can initialize the JPEG decompression object. */
    jpeg_create_decompress(&cinfo);
    
    
    /* Step 2: specify data source (eg, a file) */
    
    jpeg_stdio_src(&cinfo, infile);
    
    /* Step 3: read file parameters with jpeg_read_header() */
    
    (void) jpeg_read_header(&cinfo, TRUE);
    /* We can ignore the return value from jpeg_read_header since
    *   (a) suspension is not possible with the stdio data source, and
    *   (b) we passed TRUE to reject a tables-only JPEG file as an error.
    * See libjpeg.doc for more info.
    */
    
    /* Step 4: set parameters for decompression */
    
    /* In this example, we don't need to change any of the defaults set by
    * jpeg_read_header(), so we do nothing here.
    */
    
    /* Step 5: Start decompressor */
    
    (void) jpeg_start_decompress(&cinfo);
    /* We can ignore the return value since suspension is not possible
    * with the stdio data source.
    */
    debug("read_jpg: jpg \"%s\" is %dx%d with colormap= %d and output components= %d and actual number of colors= %d\n", 
        infilename, cinfo.output_width,cinfo.output_height,cinfo.colormap, cinfo.output_components,
        cinfo.actual_number_of_colors);
        /* We may need to do some setup of our own at this point before reading
        * the data.  After jpeg_start_decompress() we have the correct scaled
        * output image dimensions available, as well as the output colormap
        * if we asked for color quantization.
        * In this example, we need to make an output work buffer of the right size.
    */ 
    /* JSAMPLEs per row in output buffer */
    row_stride = cinfo.output_width * cinfo.output_components;
    debug("read_jpg: row_stride= %d\n", row_stride);
    /* Make a one-row-high sample array that will go away when done with image */
    buffer =  (*cinfo.mem->alloc_sarray)
        ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);
    
    // allocate texture array 
    if ((tex = (TEXTURE *)malloc(sizeof(TEXTURE))) == NULL) {
        die("read_jng: error allocating texture image");
    }
    
    tex->width = cinfo.output_width;
    tex->height = cinfo.output_height;
    tex->thumb = NULL;
    tex->tex=(GLubyte *)malloc(tex->height*tex->width*RGBA);
    debug("read_jpg: Allocated texture array OK.\n");
    
    /* Step 6: while (scan lines remain to be read) 
    /*           jpeg_read_scanlines(...);  */
    
    
    /* Here we use the library's state variable cinfo.output_scanline as the
    * loop counter, so that we don't have to keep track ourselves. */
    
    
    debug("read_jpg: output_components= %d \n", cinfo.output_components);
    debug("read_jpg: out_color_components= %d \n", cinfo.out_color_components);
    
    debug("read_jpg: tex->height= %d \n", tex->height);
    debug("read_jpg: tex->width= %d \n", tex->width);
    
    /* read image data */
    for (int i=0; i<tex->height; i++) {
        jpeg_read_scanlines(&cinfo, buffer, 1);
        for (int j=0; j<tex->width; j++) {
            for (int k = 0; k < RGBA-1; k++) {
                *(tex->tex + (RGBA*((tex->height-1-i)*tex->width+j)+k)) =
                    (GLubyte) buffer[0][(j*3)+k];
            }
            *(tex->tex + (RGBA*((tex->height-1-i)*tex->width+j)+3)) = 255;
        }		 
    }
    
    debug("read_jpg: Read image data OK \n");
    free(buffer);
    
    /* Step 7: Finish decompression */
    
    (void) jpeg_finish_decompress(&cinfo);
    debug("read_jpg: Finish decompression OK.\n");
    /* We can ignore the return value since suspension is not possible
    * with the stdio data source.
    */
    
    /* Step 8: Release JPEG decompression object */
    
    /* This is an important step since it will release a good deal of memory. */
    jpeg_destroy_decompress(&cinfo);
    
    /* After finish_decompress, we can close the input file.
    * Here we postpone it until after no more JPEG errors are possible,
    * so as to simplify the setjmp error logic above.  (Actually, I don't
    * think that jpeg_destroy can do an error exit, but why assume anything...)
    */
    fclose(infile);
    debug("read_jpg: Close input file OK.\n");
    /* At this point you may want to check to see whether any corrupt-data
    * warnings occurred (test whether jerr.pub.num_warnings is nonzero).
    */
    
    return tex;
}

//***************************************read_gif***********************************

TEXTURE* read_gif(char *infilename) {		
    int
        ImageNum = 0,
        BackGround = 0,
        OneFileFlag = FALSE,
        HelpFlag = FALSE,
        ColorMapSize = 0,
        InterlacedOffset[] = { 0, 4, 2, 1 }, /* The way Interlaced image should. */
        InterlacedJumps[] = { 8, 8, 4, 2 };    /* be read - offsets and jumps... */
    static ColorMapObject
        *ColorMap;
    TEXTURE *tex;
    int	i, j, NumFiles, Size, Row, Col, Width, Height, ExtCode, Count,
        OutFileFlag = FALSE;
    GifRecordType RecordType;
    GifByteType *Extension;
    GifRowType *ScreenBuffer;
    GifFileType *GifFile;
    GifRowType GifRow;
    static GifColorType
        *ColorMapEntry;
    
    NumFiles = 1;
    if (NumFiles == 1) {
        if ((GifFile = DGifOpenFileName((const char*)infilename)) == NULL) {
            PrintGifError();
            exit(EXIT_FAILURE);
        }
    }
    else {
        /* Use the stdin instead: */
        
        #ifdef __MSDOS__
        setmode(0, O_BINARY);
        #endif /* __MSDOS__ */
        if ((GifFile = DGifOpenFileHandle(0)) == NULL) {
            PrintGifError();
            exit(EXIT_FAILURE);
        }
    }
    debug("read_gif: Open file OK \n");
    /* Allocate the screen as vector of column of rows. We cannt allocate    */
    /* the all screen at once, as this broken minded CPU can allocate up to  */
    /* 64k at a time and our image can be bigger than that:		     */
    /* Note this screen is device independent - its the screen as defined by */
    /* the GIF file parameters itself.					     */
    if ((ScreenBuffer = (GifRowType *)
        malloc(GifFile->SHeight * sizeof(GifRowType *))) == NULL)
        die("Failed to allocate memory required, aborted.");
    
    debug("read_gif: gif \"%s\" is %dx%d\n", infilename, GifFile->SWidth, GifFile->SHeight);
    Size = GifFile->SWidth * sizeof(GifPixelType);/* Size in bytes one row.*/
    debug("read_gif: Size in bytes one row= %d\n", Size);
    
    if ((ScreenBuffer[0] = (GifRowType) malloc(Size)) == NULL) /* First row. */
        die("Failed to allocate memory required, aborted.");
    
    for (i = 0; i < GifFile->SWidth; i++)  /* Set its color to BackGround. */
        ScreenBuffer[0][i] = GifFile->SBackGroundColor;
    debug("read_gif: Background Color= %d\n", GifFile->SBackGroundColor);
    
    for (i = 1; i < GifFile->SHeight; i++) {
        /* Allocate the other rows, and set their color to background too: */
        if ((ScreenBuffer[i] = (GifRowType) malloc(Size)) == NULL)
            die("Failed to allocate memory required, aborted.");
        
        memcpy(ScreenBuffer[i], ScreenBuffer[0], Size);
    }
    debug("read_gif: Allocated row memory and set row background color OK\n");
    
    // allocate texture array 
    if ((tex = (TEXTURE *)malloc(sizeof(TEXTURE))) == NULL) {
        die("read_gif: error allocating texture image");
    }
    
    tex->width = GifFile->SWidth;
    tex->height = GifFile->SHeight;
    tex->thumb = NULL;
    tex->tex=(GLubyte *)malloc(tex->height*tex->width*RGBA);
    debug("read_gif: Allocated texture array OK.\n");
    
    /* Scan the content of the GIF file and load the image(s) in: */
    do {
        if ( DGifGetRecordType(GifFile, &RecordType) == GIF_ERROR) {
            PrintGifError();
            exit(EXIT_FAILURE);
        }
        debug("read_gif: RecordType= %d\n",RecordType);
        switch (RecordType) {
        case IMAGE_DESC_RECORD_TYPE:
            if (DGifGetImageDesc(GifFile) == GIF_ERROR) {
                PrintGifError();
                exit(EXIT_FAILURE);
            }
            Row = GifFile->Image.Top; /* Image Position relative to Screen. */
            Col = GifFile->Image.Left;
            Width = GifFile->Image.Width;
            Height = GifFile->Image.Height;
            
            if (GifFile->Image.Left + GifFile->Image.Width > GifFile->SWidth ||
                GifFile->Image.Top + GifFile->Image.Height > GifFile->SHeight) {
                debug("read_gif: Image %d is not confined to screen dimension, aborted.\n",ImageNum);
                exit(EXIT_FAILURE);
            }
            if (GifFile->Image.Interlace) {
                debug("read_gif: Interlaced image\n");
                /* Need to perform 4 passes on the images: */
                for (Count = i = 0; i < 4; i++)
                    for (j = Row + InterlacedOffset[i]; j < Row + Height;
                    j += InterlacedJumps[i]) {
                        GifQprintf("\b\b\b\b%-4d", Count++);
                        if (DGifGetLine(GifFile, &ScreenBuffer[j][Col],
                            Width) == GIF_ERROR) {
                            PrintGifError();
                            exit(EXIT_FAILURE);
                        }
                    }
            }
            else {
                debug("read_gif: Non-Interlaced image\n");
                for (i = 0; i < Height; i++) {
                    GifQprintf("\b\b\b\b%-4d", i);
                    if (DGifGetLine(GifFile, &ScreenBuffer[Row++][Col],
                        Width) == GIF_ERROR) {
                        PrintGifError();
                        exit(EXIT_FAILURE);
                    }
                }
            }
            break;
        case EXTENSION_RECORD_TYPE:
            /* Skip any extension blocks in file: */
            if (DGifGetExtension(GifFile, &ExtCode, &Extension) == GIF_ERROR) {
                PrintGifError();
                exit(EXIT_FAILURE);
            }
            while (Extension != NULL) {
                if (DGifGetExtensionNext(GifFile, &Extension) == GIF_ERROR) {
                    PrintGifError();
                    exit(EXIT_FAILURE);
                }
            }
            break;
        case TERMINATE_RECORD_TYPE:
            break;
        default:		    /* Should be traps by DGifGetRecordType. */
            break;
        }
    }
    while (RecordType != TERMINATE_RECORD_TYPE);
    
    /* Lets dump it - set the global variables required and do it: */
    BackGround = GifFile->SBackGroundColor;
    ColorMap = (GifFile->Image.ColorMap
        ? GifFile->Image.ColorMap
        : GifFile->SColorMap);
    ColorMapSize = ColorMap->ColorCount;
    
    unsigned char *Buffer, *BufferP;
    
    if ((Buffer = (unsigned char *) malloc(GifFile->SWidth * 3)) == NULL)
        die("Failed to allocate memory required, aborted.");
    
    /* read image data */
    for (i = 0; i < GifFile->SHeight; i++) {
        GifRow = ScreenBuffer[i];
        for (j = 0, BufferP = Buffer; j < GifFile->SWidth; j++) {
            ColorMapEntry = &ColorMap->Colors[GifRow[j]];
            *(tex->tex + (RGBA*((tex->height-1-i)*tex->width+j)+0)) =
                (GLubyte) ColorMapEntry->Red;
            *(tex->tex + (RGBA*((tex->height-1-i)*tex->width+j)+1)) =
                (GLubyte) ColorMapEntry->Green;
            *(tex->tex + (RGBA*((tex->height-1-i)*tex->width+j)+2)) =
                (GLubyte) ColorMapEntry->Blue;
            *(tex->tex + (RGBA*((tex->height-1-i)*tex->width+j)+3)) =
                (GLubyte) 255;						// set alpha bit
        }
    }
    debug("read_gif: Read image data OK \n");
    free(Buffer);
    
    if (DGifCloseFile(GifFile) == GIF_ERROR) {
        PrintGifError();
        exit(EXIT_FAILURE);
    }
    for (i = 0; i < GifFile->SHeight; i++)
        free(ScreenBuffer[i]);
    free(ScreenBuffer);
    debug("read_gif: Close input file OK.\n");
    
    return tex;
}

//*********************************read_bmp********************************
typedef struct BMPColorEntry {
    unsigned char	Blue;
    unsigned char	Green;
    unsigned char	Red;
    unsigned char	Reserved;
} BMPColorEntry;

void SwapLongInt (unsigned long int *L)
{ 
    *L = (*L >> 24) | ((*L >> 8) & 0x0000FF00) 
        | (*L <<24) | ((*L <<8) * 0x00FF0000);
}

void SwapShortInt (unsigned short int *S)
{ 
    *S = (*S >>8) | (*S <<8);
}

TEXTURE* read_bmp(char *infilename){
    FILE *infile;
    TEXTURE *tex;
    /*char *header, *buffer;*/
    int HEADER_SIZE = 54;
    int width, height, /*bitdepth, compression, size, planes,k.l*/ c,i,j;
    /*GLuint dataOffset,imageSize*/;
    GLubyte *image;
    int ColorTableSize/*,index*/, BytesPerPixel;
    int padding=0;
    short status;
    BMPColorEntry ColorTable[256];
    
    //Header info
    typedef struct BMPHeader {
        char type[2];
        unsigned int fileSize;
        unsigned int reserved;
        unsigned int dataOffset;
        unsigned int hdrSize;
        unsigned int width;
        unsigned int height;
        unsigned short planes;
        unsigned short bitdepth;
        unsigned int compression;
        unsigned int sizeImageData;
        unsigned int xpels;
        unsigned int ypels;
        unsigned int colorsUsed;
        unsigned int colorsImportant;
    } BMPHeader;
    
    BMPHeader BMPHdr;
    
    infile = fopen(infilename, "rb");
    if (infile==NULL)
        die("read_texture: error opening texture file %s\n", infilename);
    else
        debug("read_bmp: Open file \"%s\" OK\n",infilename);
    
    fread(BMPHdr.type, sizeof(BMPHdr.type), 1, infile);
    fread(&BMPHdr.fileSize, sizeof(unsigned int), 1, infile);
    fread(&BMPHdr.reserved, sizeof(unsigned int), 1, infile);
    fread(&BMPHdr.dataOffset, sizeof(unsigned int), 1, infile);
    fread(&BMPHdr.hdrSize, sizeof(unsigned int), 1, infile);
    fread(&BMPHdr.width, sizeof(unsigned int), 1, infile);
    fread(&BMPHdr.height, sizeof(unsigned int), 1, infile);
    fread(&BMPHdr.planes, sizeof(unsigned short), 1, infile);
    fread(&BMPHdr.bitdepth, sizeof(unsigned short), 1, infile);
    fread(&BMPHdr.compression, sizeof(unsigned int), 1, infile);
    fread(&BMPHdr.sizeImageData, sizeof(unsigned int), 1, infile);
    fread(&BMPHdr.xpels, sizeof(unsigned int), 1, infile);
    fread(&BMPHdr.ypels, sizeof(unsigned int), 1, infile);
    fread(&BMPHdr.colorsUsed, sizeof(unsigned int), 1, infile);
    fread(&BMPHdr.colorsImportant, sizeof(unsigned int), 1, infile);
    
    debug ("read_bmp: [[BMPHeader type=%c%c \nfileSize=%lu \ndataOffset=%lu \nhdrSize=%lu \nwidth=%lu \nheight=%lu \nplanes=%hu \nbitdepth=%hu "
        "\ncompression=%lu \nsizeImageData=%lu \ncolorsUsed=%lu \ncolorsImportant=%lu\n]]",
        BMPHdr.type[0], BMPHdr.type[1], BMPHdr.fileSize, BMPHdr.dataOffset, BMPHdr.hdrSize, 
        BMPHdr.width, BMPHdr.height, BMPHdr.planes, BMPHdr.bitdepth, BMPHdr.compression, 
        BMPHdr.sizeImageData, BMPHdr.colorsUsed, BMPHdr.colorsImportant);
    
    switch (BMPHdr.bitdepth)
    {
    case 1:
        debug("read_bmp: Sorry, cannot open a 1 bit monochrome BMP file.\n"); 
        debug("read_bmp: Please make another selection.\n");return NULL;
    case 4:
        ColorTableSize = 16; BytesPerPixel = 1; break;
    case 8:
        ColorTableSize = 256; BytesPerPixel = 1; break;
    case 16:
        ColorTableSize = 0; BytesPerPixel = 2; break;
    case 24:
        ColorTableSize = 0; BytesPerPixel = 3; break;
    case 32:
        ColorTableSize = 0; BytesPerPixel = 4; break;
    }
    
    switch (BMPHdr.hdrSize)
    {
    case 12:
        debug("read_bmp: Sorry, cannot open an os/2.1x BMP file.\n"); 
        debug("read_bmp: Please make another selection.\n");return NULL;
    case 240:
        debug("read_bmp: Sorry, cannot open an os/2.1x BMP file.\n"); 
        debug("read_bmp: Please make another selection.\n");return NULL;
    }
    
    switch (BMPHdr.compression)
    {
    case 1:
        debug("read_bmp: Sorry, cannot open an rle 8-bit compressed BMP file.\n"); 
        debug("read_bmp: Please make another selection.\n");return NULL;
    case 2:
        debug("read_bmp: Sorry, cannot open an rle 4-bit compressed BMP file.\n"); 
        debug("read_bmp: Please make another selection.\n");return NULL;
    case 3:
        debug("read_bmp: Sorry, cannot open a bitfields BMP file.\n"); 
        debug("read_bmp: Please make another selection.\n");return NULL;
    }
    
    // read the color pallete
    if (ColorTableSize > 0) {
        status = fread(&ColorTable,sizeof(BMPColorEntry),ColorTableSize,infile);
        if (status <=0) debug("read_bmp: Error reading BMP Color Table\n");
    }
    
    // allocate texture array 
    if ((tex = (TEXTURE *)malloc(sizeof(TEXTURE))) == NULL) {
        die("read_bmp: error allocating texture image");
    }
    width = BMPHdr.width;
    height = BMPHdr.height;
    tex->width = width;
    tex->height = height;
    tex->thumb = NULL;
    tex->tex=(GLubyte *)malloc(width*height*RGBA);
    debug("read_bmp: Allocated texture array OK.\n");
    
    if (BMPHdr.bitdepth == 4) 	// 4 bit - 16 color
        padding = (4 - (((int)(ceil(width/2.0)) * BytesPerPixel) % 4)) % 4;
    else
        padding = (4 - ((width * BytesPerPixel) % 4)) % 4;
    debug("read_bmp: read_bmp: padding= %d\n",padding);
    
    if (BMPHdr.sizeImageData == 0)
        BMPHdr.sizeImageData=width*height*BytesPerPixel+padding*height;
    
    image = (GLubyte *)malloc(BMPHdr.sizeImageData);
    if (image == NULL) die("read_bmp: error mallocing buffer\n");
    debug("read_bmp: new BMPHdr.sizeImageData= %d\n",BMPHdr.sizeImageData);
    
    //read the pixel data
    fseek(infile, BMPHdr.dataOffset, SEEK_SET);
    
    fread(image,sizeof(unsigned char),BMPHdr.sizeImageData,infile);
    
    debug("read_bmp: BytesPerPixel= %d\n",BytesPerPixel);
    
    int r, g, b;
    short cw;
    int chi, clo;
    c = 0, i = height;
    while ( i > 0 ) {
        i--;
        for (j = 0; j < width; j++ ) {
            
            if (BMPHdr.bitdepth == 4) {	// 4 bit - 16 color
                chi = image[c]>>4;	
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+0)) = 
                    ColorTable[chi].Red;
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+1)) = 
                    ColorTable[chi].Green;
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+2)) = 
                    ColorTable[chi].Blue;
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+3)) = 255;
                
                j++;
                clo = (image[c]&0x0F);	
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+0)) = r =
                    ColorTable[clo].Red;
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+1)) = g =
                    ColorTable[clo].Green;
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+2)) = b =
                    ColorTable[clo].Blue;
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+3)) = 255;
                
                c+=1;
            }
            if (BMPHdr.bitdepth == 8) {	// 8 bit - 256 color
                
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+0)) = r =
                    ColorTable[image[c]].Red;
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+1)) = g =
                    ColorTable[image[c]].Green;
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+2)) = b =
                    ColorTable[image[c]].Blue;
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+3)) = 255;
                
                c+=1;
            }
            if (BMPHdr.bitdepth == 16) {	// 16 bit - high color
                
                cw = image[c] + (((short)image[c+1])<<8);
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+0)) = 
                    (GLubyte) (((cw&0x7C00)>>10)<<3);
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+1)) = 
                    (GLubyte) (((cw&0x03E0)>>5)<<3);
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+2)) = 
                    (GLubyte) (((cw&0x1F))<<3);
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+3)) = 255;
                
                c+=2;
            }
            if (BMPHdr.bitdepth == 24) {	// 24 bit - true color
                
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+0)) = 
                    (GLubyte) image[c+2];
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+1)) =
                    (GLubyte) image[c+1];
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+2)) =
                    (GLubyte) image[c];
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+3)) = 255;
                
                c+=3;
            }
            if (BMPHdr.bitdepth == 32) {	// 32 bit - true color
                
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+0)) = 
                    (GLubyte) image[c+2];
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+1)) =
                    (GLubyte) image[c+1];
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+2)) =
                    (GLubyte) image[c];
                *(tex->tex + (RGBA*((height-1-i)*(width)+j)+3)) = 
                    (GLubyte) image[c+3];
                c+=4;
            } 
        } // end for
        c += padding;
    } // end while
    fclose(infile);
    free(image);
    return tex;
}
