/*
 * License for Tiled Display modifications
 *
 * Copyright (c) 2005 Mohamed Ibrahim and Stephanie Smullen.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include an acknowlegement that the software contains
 *    code based on contributions made by Mohamed Ibrahim and Stephanie Smullen and
 *    include a link to http://zog.utc.edu/~vislab/.
 *
 * 4.  The Original code contains third party software libraries copyrighted 
 *	under their own licenses. Redistribution of Libraries in source code, 
 *	with or without modification, or redistribution in binary form falls
 *	under the terms of the licenses for the corresponding Libraries. 
 *
 * This software is provided "as is" and any expressed or implied
 * warranties, including, but not limited to, the implied warranties
 * of merchantability and fitness for a particular purpose are
 * disclaimed.  In no event shall the authors or their
 * employers be liable for any direct, indirect, incidental,
 * special, exemplary, or consequential damages (including, but not
 * limited to, procurement of substitute goods or services; loss of
 * use, data, or profits; or business interruption) however caused and
 * on any theory of liability, whether in contract, strict liability,
 * or tort (including negligence or otherwise) arising in any way out
 * of the use of this software, even if advised of the possibility of
 * such damage.
 *

 *
 */
#include <windows.h>
#include <winsock.h>
#include "network_comm.h"

static SOCKET the_socket;

int initNetwork(int port, char* server){
	WORD sock_version;
	WSADATA wsa_data;
	int nret;
	sock_version = MAKEWORD(1, 1);
	WSAStartup(sock_version, &wsa_data);
	LPHOSTENT host_entry;
	host_entry = gethostbyname(server);	
	if (!host_entry) {
		nret = WSAGetLastError();		
		WSACleanup();
		return NETWORK_ERROR;
	}
	the_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (the_socket == INVALID_SOCKET) {
		nret = WSAGetLastError();
		WSACleanup();
		return NETWORK_ERROR;
	}
	SOCKADDR_IN serverInfo;
	serverInfo.sin_family = AF_INET;
	serverInfo.sin_addr = *((LPIN_ADDR)*host_entry->h_addr_list);
	serverInfo.sin_port = htons(port);
	nret = connect(the_socket, (LPSOCKADDR)&serverInfo, sizeof(struct sockaddr));
	if (nret == SOCKET_ERROR) {
		nret = WSAGetLastError();
		WSACleanup();
		return NETWORK_ERROR;
	}
	return NETWORK_OK;
}

void finiNetwork(){
	closesocket(the_socket);
	WSACleanup();
}

int sendNetwork(char* buffer){
	if(strlen(buffer) > NET_BUFFER_SIZE) return NETWORK_ERROR;	
	if (send(the_socket, buffer, strlen(buffer), 0) == SOCKET_ERROR) {
		return NETWORK_ERROR;
	}
	return NETWORK_OK;
}

int recvNetwork(char* buffer){
	int nret = recv(the_socket, buffer, NET_BUFFER_SIZE, 0);
	if (nret == SOCKET_ERROR || nret > NET_BUFFER_SIZE) {
		buffer[0] = 0;
		return NETWORK_ERROR;
	} 
	buffer[nret] = 0;
	return NETWORK_OK;
}
